

#  A mud client for
#  Dawn of the Ages  http://dawn.rplay.net
#
#  Copyright (C) 2001,2002 Peter Herth <p.herth@uni-koeln.de>
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 

import sys
import telnetlib
import thread
import types

import time
import string
import traceback
import os
import os.path
import re
import struct
import copy
import sched
import threading
#sys.setdefaultencoding = "iso-8859-1"

have_readline = None
have_signal = None
have_fcntl = None
have_termios = None


try:
    import readline
    have_readline = 1
except:
    pass
    ##print "Module readline not found."

try:
    import signal
    have_signal = 1
except:
    pass
    ##print "Module signal not found."


try:
    import fcntl
    have_fcntl = 1
except:
    pass
    #print "Module fcntl not found."


#import TERMIOS
try:
    import termios
    have_termios = 1
except:
    pass
    ##print "Module termios not found."

have_tkinter = None

try:
    from Tkinter import *
    from tkMessageBox import *
    from tkSimpleDialog import *
    have_tkinter = 1
except:
    print "Module Tkinter not found."
    pass


if have_tkinter:
    import Gui
    import Term
    import Clock

try:
    from Utilities import *
except:
    pass

debug = 0
client = None


#variables used for configuration


tf_look = None

width = 80
height = 25

buffer = ""

lasttime = 0
done = None

last_line = ""

splitscreen = 0
split_line=20
status_size = 3
input_size = 3

eval_brackets = 1
fifo = [None]*20
input_row = 1


more = None

more_buffer = []

recall_buffer = []

#echo_commands = 0   #print commands sent to mud in the output screen also

actives = {}
fadings = {}
affects = {}
std_affects = []
args = sys.argv
if len(args)> 1:
    if args[1]=="--tf":
        tf_look = 1
        if len(args)>2:
            args = args[0:1]+args[2:]
        else:
            args = args[0:1]


if tf_look:
    status_size = 1
    input_size = 3

t = None # telnet socket
#recall_size = 10000

#list of save-functions that are called by #write automatically.
global save_functions

# options for the spawned xterms - feel free to overwrite them for your own preferences
#xterm_opts = '-bg black -fg grey -sb -rightbar -sl 10000 -font "-adobe-courier-bold-r-normal-*-17-120-100-100-m-100-iso8859-*"'
xterm_opts = '-bg black -fg grey -sb -rightbar -sl 10000 '

#prompt = "<%h/%Hhp %m/%Mm %v/%Vmv Exits:%e!%Xtnl T:%Y F:%F!%i!>prompt>"
infograbber_prompt = "<%h/%Hhp %m/%Mm %v/%Vmv Exits:%e!%Xtnl T:%Y F:%F!%i!>prompt>"

#brackets = re.compile(r"[^\\]\$\[(.*?)\]", re.MULTILINE)
brackets = re.compile(r"\$\[(.*?)\]", re.MULTILINE)

#regexp for variable-substitution
#This means a single '$' followed by a number
VARIABLE_PATTERN=re.compile("([^$]+|^)\$([0-9])")

#autologin stuff
name_prompt_re = re.compile(r"How shall thy name")
password_prompt_re = re.compile(r"^ Password:")
enter_re = re.compile(r"Now you will enter|Welcome back")

#clearup = "\33c\33[2J"
clearup = "\33[r\33[2J"

SAVEPOS = "\33[s" #+"7"
RESTOREPOS = "\33[u" #+"8"
SAVEPOS = "\33"+"7"
RESTOREPOS = "\33"+"8"

#ESC [ 1K        Erase from beginning of line to cursor

CLR_TO_CURSOR="\033[1K"


CLS  = "\033[2J"     # clear to bottom of screen
HOME = "\033[0;0H"   # cursor home

# color code number definition (0-7)
(BLACK,RED,GREEN,YELLOW,BLUE,MAGENTA,CYAN,WHITE)=range(8)


def gotoXY(x,y):
    return "\33[%d;%dH" % (y,x)

def fcolor(c,bold=0,dim=0,reverse=0,underline=0):
    b =""
    if bold:
        b ="1;"
    d = ""
    if dim:
        d="2;"
    r = ""
    if reverse:
        r="7;"
    u=""
    if underline:
        u="4"
    return "\33[%s%s%s%s3%dm" % (b, d, r, u, c)
def bcolor(c,bold=0,dim=0,reverse=0):
    b =""
    if bold:
        b ="1;"
    d = ""
    if dim:
        d="2;"
    r = ""
    if reverse:
        r="7;"
    return "\33[%s%s%s4%dm" % (b, d, r, c)

#create ansi escape sequences

def bold(on=1):
    if on:
        return "\33[1m"
    else:
        return ""
def dim(on=1):
    if on:
        return "\33[2m"
    else:
        return ""
def reverse(on=1):
    if on:
        return "\33[7m"
    else:
        return ""

def underline(on=1):
    if on:
        return "\033[4m"
    else:
        return ""
    

#reset colors
def clr():
    return "\33[0m"

CLR = clr()

# foreground color
def fg(n):
    return "\33[3%dm" % n

# background color
def bg(n):
    return "\33[4%dm" % n

statuscolor = dim(1)+reverse(1)+fg(WHITE)+bg(BLACK)

# uncomment this for change of status color, in this case, black
# it can be changed at runtime by:
# #eval statuscolor = bg(RED)
# #split

statuscolor = fg(WHITE)+bg(BLACK)

echocolor=""


colors = { "red"     : fcolor(RED),
           "blue"    : fcolor(BLUE),
           "white"   : fcolor(WHITE),
           "magenta" : fcolor(MAGENTA),
           "cyan"    : fcolor(CYAN),
           "yellow"  : fcolor(YELLOW),
           "green"   : fcolor(GREEN),
           "black"   : fcolor(BLACK),
           "Red"     : fcolor(RED, bold=1),
           "Blue"    : fcolor(BLUE, bold=1),
           "White"   : fcolor(WHITE, bold=1),
           "Magenta" : fcolor(MAGENTA, bold=1),
           "Cyan"    : fcolor(CYAN, bold=1),
           "Yellow"  : fcolor(YELLOW, bold=1),
           "Green"   : fcolor(GREEN, bold=1),
           "Black"   : fcolor(BLACK, bold=1),
           "bold" : bold(),
           "dim": dim(),
           "reverse": reverse(),
           "underline": underline()
           }

def parse_color(text):
    """parses text into a colorcode
    """
    if colors.has_key(text):
        return colors[text]
    else:
        return ""
    
# show a table of all creatable color settings :)

def colortable(client):
    client.output("\n")

    for b in range(2):
        for d in range(2):
            for r in range (2):
                txt = ("bold:%d dim:%d reverse:%d\nf|b>" % (b,d,r))
                for ba in range(8):
                    txt = txt + ("  %d " % ba)
                txt = txt +  "\n"
                for f in range(8):
                    txt = txt + ("%d   " % f)
                    for ba in range(8):
                        txt = txt + (bold(b)+dim(d)+reverse(r)+fg(f)+bg(ba)+"XXXX"+clr())
                    txt = txt + ("\n")
                txt = txt + ("\n")
                client.output(txt)
    client.output("\n")


def evaluate_brackets(text):
    match = brackets.search(text)
    while match:
        ##print "match:"+match.group(1)+"<="
        newtext=match.group(1)
        try:
            newtext = "%s" % eval(match.group(1))
        except:
            pass
        text = text[:match.start(1)-2]+newtext+text[match.end(1)+1:]        
        match = brackets.search(text)
    return text

# write "txt" at the coordinates (x,y) in the status bar y=[1,2], x=[1..80]
def write_status(x,y,txt):
    if splitscreen:
        ##sys.stdout.write("\33[%d;%dH%s\33[%d;%dH" % (split_line+y-1,x, txt, split_line+2, 1))
        sys.stdout.write(SAVEPOS+gotoXY(x,split_line+y-1)+txt+RESTOREPOS)
        

def write_to_screen(buffer):
    if splitscreen:
        ##sys.stdout.write("\33[6n")
        #sys.stdout.write(("\33"+"8\r\n%s\33"+"7\33[%d;%dH") % (buffer, split_line+2, input_row))
        sys.stdout.write("\33[0m"+SAVEPOS+gotoXY(1,split_line-1)+"\33[0m\33E"+buffer+RESTOREPOS)
    else:
        sys.stdout.write("\n")
        sys.stdout.write(buffer)
    sys.stdout.flush()


def more_display():
    if more:
        return "(MORE %d)" % len(more_buffer)
    return ""
    #return "         "


# create an xterm listening to the fifo nr
def open_fifo(nr,title="Mud Client"):
    fifoname = "/tmp/mudc_fifo_%d" % nr
    if not os.path.exists(fifoname):
        if not client.get_option("quiet"):
            client.output("creating:"+fifoname+"\n")
        os.mkfifo(fifoname)
    term = os.popen("xterm %s -T '%s' -e cat %s &" % (xterm_opts, title, fifoname))
    fifo[nr]=(open(fifoname,"w"),term)

def close_fifo(nr):
    if fifo[nr]:
        fifo[nr][0].close()
        fifo[nr][1].close()
        fifo[nr] = None
    
def send_fifo(nr,text):
    if fifo[nr]:
        fifo[nr][0].write(text)
        fifo[nr][0].flush()

# return expansion of text with the given match set
# eg:  asdf $1 qwert $2 => asdf (matchgroup1) qwert (matchgroup2)
## def expand(text, matches):
##     try:
##         #print "expand t:",text,"matches:",matches
##         if type(matches)==type(" "):
##             matches =(matches,)
##         ##print "expand:", text, matches
##         pos = string.find(text,"$")
##         ##print "pos:", pos
##         while pos >= 0:
##             nr = -1
##             try:
##                 nr = int(text[pos+1])
##             except:
##                 nr = -1
##             if nr == 0:
##                 t = string.join(matches," ")
##                 text = text[:pos]+t+text[pos+2:]
##             elif nr >0 and nr <= len(matches):
##                 text = text[:pos]+matches[nr-1]+text[pos+2:]
##             ##else: 
##             ##    text = text[:pos]+text[pos+2:]
##             pos = string.find(text,"$",pos+2)
##             #print "erg:", text, matches
##             #print "pos:", pos
##         ##print "result:",text
##     except:
##         pass
##     return text
def expand(text,matches):
    #replaces unescaped variables with the parameters in matches
    global VARIABLE_PATTERN
    if type(matches)==type(" "):
        matches =(matches,)
    def replace_variables(match):
        #try:
            #group(1) is the text left of the '$'
            #group(2) is the number following '$'
            nr = int(match.group(2))
            if nr ==0: #catch $0
                return match.group(1)+' '.join(matches)
            else:
                return match.group(1)+matches[int(match.group(2))-1]
        #except:
        #    pass
        
    text = VARIABLE_PATTERN.sub(replace_variables,text)
    return text.replace("$$","$")

# list of all applets 
gags = []
highlights = []
applets = []

# list of all displayers

displayers = []

# Base class of all applet, that is a trigger defined by a regexp,
# that runs a function on matching the input stream, and possibly
# displays a result in the tatusbar, can gag text in the process

# any user extending the class only has to overwrite the constructor
# when  appropriate and define match_fun to do the actual work,
# filling self.display (which gets painted in the status area) and
# returning a result (for gagging and reformatting), returning nothing
# or None results in no change of the input text

class Applet:
    def __init__(self, trigger_re, x=None, y=None, re_params = re.MULTILINE):   # applet is only displayed when x and y > 0
        self.trigger = re.compile(trigger_re, re_params)
        self.x=x
        self.y=y
        self.display = ""
        
    def match_fun(self, match, text):
        pass

    def do_fa_match(self, text): #unused
        l = self.trigger.findall(text)
        for m in l:
            if type(m)==type(" "):
                m =(m,)
            result = self.match_fun(m, text)
            if self.x and self.y:
                write_status(self.x,self.y,self.display)
            if result:
                return result
        return text
                    

    # the function driving the functionality
    def do_match(self,text): 
        pos = 0                               # startposition for matching
        #print "matching",self.trigger,"at",pos
        match = self.trigger.search(text,pos) # try to get a match
        while match:                          # for all matches found in input
            delta = 0
            result = self.match_fun(match, text) # get replacement text for match
            if self.x and self.y:                # display output, when wanted
                write_status(self.x,self.y,self.display)
            if result != None:                           # if there is a result returned
                ##send_fifo(1,"matchresult:"+result+"<=\n")
                matchlen = match.end()-match.start()  # get length of match
                delta = matchlen - len(result)        # and compare with len of result
                #print "result:"+result+"<="
                newtext = text[:match.start()]+result+text[match.end():] # replace match with result

                #print "gag:"+text+"->"+newtext+"<-"
                text = newtext
            #print "matching",self.trigger,"at",pos
            match = self.trigger.search(text, match.end()-delta)  # try to get match in rest text
        return text # return edited text

class Affect:
    def __init__(self, name, symbol, aff_re, spellup, spelldown, fading=None,standard = 1):
        self.name = name
        self.symbol = symbol
        self.set_aff_re(aff_re)
        self.standard = standard
        #print "restring:",re_string
        self.spellup = re.compile(spellup, re.MULTILINE)
        self.spelldown = re.compile(spelldown, re.MULTILINE)
        if fading:
            self.fading = re.compile(fading, re.MULTILINE)
        else:
            self.fading = None

    def set_aff_re(self,aff_re):
        self.aff_re_orig=aff_re
        re_string = r"^\s*"+aff_re+r"\s*$|^\s*"+aff_re+r"\s*\:"
        self.aff_re = re.compile(re_string,re.MULTILINE)
        
    def __repr__(self):
        return "<Affect:"+self.name+">"
 
    def do_match(self,buffer):
        if self.spellup.search(buffer):
            actives[self.name]=self.symbol
 
        if self.aff_re.search(buffer):
            actives[self.name]=self.symbol
 
        if self.spelldown.search(buffer):
            if actives.has_key(self.name):
                del actives[self.name]
                if fadings.has_key(self.name):
                    del fadings[self.name]
                update_statusbar()
                
        if self.fading:
            if self.fading.search(buffer):
                actives[self.name]=reverse()+self.symbol
                fadings[self.name]=self.symbol

       
class Gag(Applet):
    def __init__(self, trigger_re, replace="", alternate=None, alternate_exp=None, re_opts=re.MULTILINE):
        self.trigger = re.compile(trigger_re, re_opts)
        self.replace = replace
        self.alternate = alternate
        self.alternate_exp = alternate_exp
        self.trigger_re = trigger_re
        self.x = None
        self.y = None
        
    def match_fun(self, match, text):
        g = match.groups()
        #print "gagmatch:",g
        erg = expand(self.replace, g)
        
        if self.alternate:
            self.alternate.output(expand(self.alternate_exp, g)+"\n")
            #send_fifo(self.alternate, expand(self.alternate_exp, g)+"\n")

        return erg

#gags.append(Gag(r"LOG:(.*)$", "", 2, "LOG: $1"))


class HighLight(Applet):
    def __init__(self, trigger_re, colorspec=""):
        self.trigger_re = trigger_re
        self.trigger = re.compile("("+trigger_re+")", re.MULTILINE)
        self.colorspec = colorspec
        self.color = parse_color(colorspec)


    
    def do_match(self, text):
        #print "hl:",text
#        m = self.trigger.search(text)
#        if m:
#            send_fifo(1,"=>%s<=\n" % m.groups())
        
        text = re.subn(self.trigger,self.color+r"\1"+CLR,text)[0]
        
#        text = re.subn(self.r2, CLR+self.color,text)[0]

        #print "hl result:", text
        return text
    
#highlights.append(HighLight(r"(claw)", fcolor(BLUE,bold=1)))
#highlights.append(HighLight(r"^(.*sphere.*)$", "\033[4m"))


def update_statusbar():
    for d in displayers:
        write_status(d[0],d[1],d[2]())
# @param keycode: keycode for non-posix oses
class keybinding:
    def __init__(self, client, keydef, command,keycode=None):
        self.keydef = keydef
        self.command = command
        self.client = client
        self.keycode=keycode
        
    def bind(self):
        shift=""
        ctrl=""
        alt=""
        code = self.keydef
        if code[:5]=="ctrl-":
            ctrl="Control-"
            code = code[5:]
        if code[:6]=="shift-":
            shift="Shift-"
            code = code[6:]
        if code[:4]=="Alt-":
            alt="Alt-"
            code = code[4:]
        if (self.keycode and (os.name != "posix")) or not self.keycode:
            #if a keycode is defined, define the binding under windows only
            if self.client.gui:
                try:
                    self.client.gui.bind_all("<%s%s%sKeyPress-%s>" % (ctrl, shift, alt, code), self.do_command)
                    self.client.gui.entry.bind("<%s%s%sKeyPress-%s>" % (ctrl, shift, alt, code), self.do_command)
                    #if we do not override the standard-bindings, we will add numbers to the entry-field when numpad-walking
                    
                except:
                    self.client.output("cannot bind key %s" % self.keydef)

    def do_command(self, event):
    #    print event.keycode
        if (not self.keycode) or (event.keycode == self.keycode):
            #run bindings without a keycode defined under all oses
            self.client.process_command(self.command)
            return "break"
    
################### implementation of the commands #################

def tab_pressed(client, text):
    global more
    global more_buffer

    if (len(more_buffer)>0):
        n = min(20, len(more_buffer))
        txt = string.join(more_buffer[:n],"\n")
        more_buffer = more_buffer[n:]
        client.raw_output(txt)
        if client.gui:
            client.gui.master.title("%s [%d]"% (client.title, len(more_buffer)))
        update_statusbar()



def do_more(client, param):
    """ Lets the output stop after each screenful
    usage: #more
    The TAB-key displays the next screenful
    """
    global more
    global more_buffer
    if more:
        txt = string.join(more_buffer,"\n")
        more_buffer = []
        client.raw_output(txt)
        more = None
        client.output("More mode cancelled.\n")
	if client.gui:
	    client.gui.master.title(client.title)
        if have_readline:
            readline.parse_and_bind('TAB: complete')
    else:
        client.output("More modus started\n")
        more = 1
        if have_readline:
            readline.parse_and_bind('TAB: "~tab_pressed\n"')
    #snl()
    
def do_delay(client,params):
    """ issues delayed commands
    usage #delay <time> <commands to use, separated by \\;>

    <time> gives the delay between each command in milliseconds.
    a time of -1 means the commands will be run ""synchronized"",
    i.e. each time a prompt is recieved from the MUD the next one
    is issued.
    Commands queued in this fashion can be stopped with #break
    """
    if params:
        #try:
            p = params.split(" ",1)

            if p[0] == "-1":
                client.append_queue(p[1])
            else:
                client.output("Added as job no. %i\n"%client.delay(float(p[0]),p[1]))
                
        #except:
        #    client.output("usage: #delay <time> <commands to use, separated by \\;>")
    else:
        client.output("usage: #delay <time> <commands to use, separated by \\;>")   
    
        
def do_break(client,params):
    """ deletes delayed commands
    usage #break [-d [<nr>]|-s]
    Without arguments: deletes all synchronized and/oror delayed commands.
    -s: remove all synchronized commands (from #delay -1 )
    -d: remove all delayed jobs
    -d <number>: stop job <number> of the delayed jobs
    """
    if params:
        p = params.split(" ")
        if p[0] == "-s":
            client.clear_queue()
        elif p[0] == "-d":
            if len(p)>1:
                nr = int(p[1])
                client.cancel_delay(nr)
            else:
                client.cancel_delay()
    else:
        client.clear_queue()
        client.cancel_delay()
        
def list_jobs(client,params):
    """shows active delayed jobs
    usage #jobs
    shows jobs delayed with ##delay
    """
    i = 0
    client.output("Delayed jobs:")
    client.output("Nr.  Delay (ms)  First command")
    for t in client.delays:
        if t.isAlive():
            client.output("%3d %8d %15s"%(i,t.time,t.commands[0]))
        i = i + 1
    
def dorepeat(client,params):
    """ repeats a command a certain ammount of times
    usage #repeat [-d <time>] <count> <commands to use, separated by \\;>
    <time> gives the delay between each command in milliseconds.
    a time of -1 means the commands will be run ""synchronized"",
    i.e. each time a prompt is recieved from the MUD the next one
    is issued.
    Commands queued in this fashion can be stopped with #break
    """
    if params:
        p=string.split(params," ",1)
        synchronized = 0
        time = 0
        while p[0][0] == "-":
            if p[0][1] == "d":
                p = string.split(p[1]," ",1)
                time = int(p[0])
            params=p[1]
            p = string.split(params," ",1)
        try:
            count = int(p[0])
            if time == -1:
                commands = []
                for i in range(count):
                    commands.append(p[1])
                client.append_queue(commands)
            elif time:
                commands = []
                for i in range(count):
                    commands.append(p[1])
                client.output("Added as job no. %i\n"%client.delay(float(time),commands))
            else:
                for i in range(count):
                    client.process_command(p[1])
        except:
            client.output("Syntax: #repeat [-s|-d <time>] <count> <commands, separated by \\;>")

def doif(client,params):
    """ execute commands based on conditions
    usage: #if $[expr] <commands to be executed if expr is true>
    $[expression] must be a valid python expression and enclosed in $[...]
    The easiest would be a test on a variable set with #setvar:
    #setvar test ""test""

    #if $[test == ""test""] #echo hi
    will execute #echo hi
    #if $[test == ""lo""] #echo hi
    will execute nothing
    """
    if params:
        match = brackets.split(" "+params,1)
        if match:
            if eval(match[1]):
                client.process_command(string.strip(match[2]))
                                      #else:
        #    client.process_command(p[2])
            
    else:
        client.output("usage: #if $[expression] <commands,separated by \;>")
            

def resetaffects(client, params):
    """ resets affects 
    usage #resetaffects
    all affects will be reset to the original values,
    custom affects will be deleted
    """
    client.resetaffects()
    if not client.get_ioption("quiet"):
        client.output("Ok")
            
def defaffect(client, params):
    """ manipulates and displays affects for the statusbar
    usage #affect <name>~<symbol>~<affect>~<spellup>~<spelldown>[~<fade>]
          creates or overwrites an affect
    name: Name of the affect
    symbol: How the affect will be displayed in the statusbar, use Dawn-style
            colorcodes
    affect: how the affect appears in the list from the affect-command
    spellup: message from the MUD when the affect starts
    spelldown: message from the MUD when the affect ends
    fade: optional, message on last tick before the affect ends
    usage #affect <name>
          displays all information to the specified affect
    usage #affect
          lists names and symbols of all defined affects
    """
    if len(params) > 1:
        try:
            p=string.split(params,"~")
            print len(p)
            if len(p) >1:
                name = p[0]
                symbol = p[1]
                aff_re = p[2]
                spellup = p[3]
                spelldown = p[4]
                fade = None
                if len(p) > 5:
                    fade = p[5]
                aff = None
                if client.get_affects().has_key(name):
                    aff = a
                if aff:
                    if not client.get_ioption("quiet"):
                        client.output("Overwriting affect %s."%name)
                    aff.symbol=stringtoansi(symbol)
                    aff.set_aff_re(aff_re)
                    aff.spellup=re.compile(spellup,re.MULTILINE)
                    aff.spelldown=re.compile(spelldown,re.MULTILINE)
                    aff.standard = 0
                    if fade:
                        aff.fading = re.compile(fade,re.MULTILINE)
                    else:
                        aff.fading = None
                else:
                    if not client.get_ioption("quiet"):
                        client.output("Defined new affect %s."%name)
                    aff = Affect(name,symbol,aff_re,spellup,spelldown,fade,standard=0)
                    affects[aff.name]=aff
            else:
                aff = None
                name = p[0]
                if client.get_affects().has_key(name):
                    aff = client.get_affects()[name]
                if aff:
                    client.output("Affect %s is defined as follows:"%aff.name)
                    client.output("Name: %s \nSymbol: %s \nColorcode for Symbol: %s"%(aff.name, aff.symbol, ansitostring(aff.symbol)))
                    client.output("Name in affectlist: "+aff.aff_re_orig)
                    client.output("Spellup: "+aff.spellup.pattern)
                    client.output("Spelldown: "+aff.spelldown.pattern)
                    if aff.fading:
                        client.output("Fading message: %s"%aff.fading.pattern)
                else:
                    client.output("No such affect found")
                    
                        
        except:
            client.output("usage #affect <name>~<symbol>~<affect>~<spellup>~<spelldown>[~<fade>]\n")
    else:
        client.output("The following affects are defined:\n")
        client.output("Name                    |Symbol|")
        for a in client.get_affects().values():
            #print a
            client.output("%-25s|%-6s"%(a.name,a.symbol)) 
                                
def delaffect(client,params):
    """ removes an affect
    usage #delaffect <name>
    Note:
    Predefined affects will be restored at each startup!"""
    if params:
        if client.get_affects().has_key(params):
            client.get_affects().remove(params)
        else:
            client.output("No such affect: "+params)
                
    else:
        client.output("usage: #delaffect <name>")
        
                
                
def defbutton(client, params):
    """ defines a spellbutton for the gui (gui)
    usage #defbutton <name>~<commands to use, separated by \\;>
    name: name of the button (if the name of an affect is chosen,
          it will automatically switch on and off)
    """
    if params:
        param = params.split('~')
        if client.gui:
            client.gui.create_affect_button(param[0],param[1])

def buttoncolor(client,params):
    """ sets the colors for spellbuttons (gui)
    usage #buttoncolor <status> <background|foreground> <color>
    
    status: can be active (affect present), fading (last tick) or inactive (affect not present)
    background/foreground: which color to set
    Possible colors are:
    black, white, gray, green, blue, lightblue, etc.
    Alternatively, the color can be given as RGB-values, separated by ','.
    Eg: #buttonfg custom~128,128,128
    """
    try:
        if params:
            if client.gui:
                p=params.split(" ",3)
                try:
                    color=client.gui.winfo_rgb(p[2])
                except:
                    client.output("wrong color definition: "+p[2])
                    return
                if p[0]=="fading":
                    if p[1]=="background":
                        client.gui.affect_button_fading_bg=p[2]
                    elif p[1]=="foreground":
                        client.gui.affect_button_fading_fg=p[2]
                    else:
                        client.output("Second parameter must be ""background"" or ""foreground""")
                        client.output("usage: #buttoncolor <status> <background|foreground> <color>")
                elif p[0]=="active":
                    if p[1]=="background":
                        client.gui.affect_button_active_bg=p[2]
                    elif p[1]=="foreground":
                        client.gui.affect_button_active_fg=p[2]
                    else:
                        client.output("Second parameter must be ""background"" or ""foreground""")
                        client.output("usage: #buttoncolor <status> <background|foreground> <color>")
                elif p[0]=="inactive":
                    if p[1]=="background":
                        client.gui.affect_button_inactive_bg=p[2]
                    elif p[1]=="foreground":
                        client.gui.affect_button_inactive_fg=p[2]
                    else:
                        client.output("Second parameter must be ""background"" or ""foreground""")
                        client.output("usage: #buttoncolor <status> <background|foreground> <color>")
                else:
                    client.output("<status> must be ""active"" or ""fading"" or ""inactive""")
                client.gui.color_buttons()
            else:
                client.output("Gui must be active for spellbuttons.")
        else:
            client.output("usage #buttoncolor <status> <background|foreground> <color>")
    except:
        client.output("usage #buttoncolor <status> <background|foreground> <color>")
            
            
                
def buttonbg(client, params):
    """ sets the background-color for a spellbutton. (gui)
    usage #buttonbg <name>~<color>
    name : name of the button
    If the button is named after an affect, the setting will only be
    temporary.
    Possible colors are:
    black, white, gray, green, blue, lightblue, etc.
    Alternatively, the color can be given as RGB-values, separated by ','.
    Eg: #buttonfg custom~128,128,128
    """
    if params:
        if client.gui:
            p=params.split("~",1)
            but = None
            for button in client.gui.affect_buttons:
                if button.name == p[0]:
                    but = button
            if but:
                try:
                    but.configure(background=p[1])
                except:
                    try:
                        col = p[1].split(",",3)
                        col = (int(col[0]),int(col[1]),int(col[2]))
                        but.configure(background="#%02x%02x%02x"%col)
                    except:
                        client.output("Wrong color: "+p[1])
            else:
                client.output("No button defined with name: "+p[0])
        else:
            client.output("You need to have the gui active for spellbuttons to work.")
    else:
        client.output("usage #setbutton <name>~<on|fading|off>")
        
def buttonfg(client, params):
    """ sets the foreground-color for a spellbutton. (gui)
    usage #buttonbg <name>~<color>
    name : name of the button
    Possible colors are:
    Possible colors are:
    black, white, gray, green, blue, lightblue, etc.
    Alternatively, the color can be given as RGB-values, separated by ','.
    Eg: #buttonfg custom~100,0,200
    
    """
    if params:
        if client.gui:
            p=params.split("~",1)
            but = None
            for button in client.gui.affect_buttons:
                if button.name == p[0]:
                    but = button
                    
            if but:
                try:
                    but.configure(foreground=p[1])
                except:
                    try:
                        col = p[1].split(",",3)
                        col = (int(col[0]),int(col[1]),int(col[2]))
                        but.configure(background="#%02x%02x%02x"%col)
                    except:
                        client.output("Wrong color: "+p[1])
            else:
                client.output("No button defined with name: "+p[0])
        else:
            client.output("You need to have the gui active for spellbuttons to work.")
    else:
        client.output("usage #buttonfg <name>~<on|fading|off>")

def buttoncmd(client, params):
    """ sets the commands for a spellbutton. (gui)
    usage #buttonbg <name>~<commands, separated by \;>
    name : name of the button
    """
    if params:
        if client.gui:
            p=params.split("~",1)
            but = None
            for button in client.gui.affect_buttons:
                if button.name == p[0]:
                    but = button
                    
            if but:
                but.command=p[1]
                client.output("Button "+but.name+" now invokes: "+p[1])
            else:
                client.output("No button defined with name: "+p[0])
        else:
            client.output("You need to have the gui active for spellbuttons to work.")
    else:
        client.output("usage #buttoncmd <name>~<commands, separated by \;>")
       
    
                    
def rembutton(client, params):
    """ removes a spellbutton from the gui (gui)
    usage #rembutton <name>
    """
    if params:
       if client.gui:
           client.gui.remove_affect_button(params)
           
def defalias(client, text):
    """    defines an alias or lists the defined ones
    usage: #alias <shortcut> <commands to use, separated by \\; >
    You can escape the ';' by '\\', eg:
    #alias rw get $1\\\\;wear $1
    """
    
    if len(text)<1:
        client.output("\nThe following aliases are defined:\n")
        keys = client.aliases.keys()
        keys.sort()
        for k in keys:
            client.output(k+" aliases "+client.aliases[k]+"\n")
        if len(keys)<1:
            client.output("None.")
        #snl()
        return
    (aliasname,aliastext)= string.split(text," ",1)
    client.aliases[aliasname]=aliastext
    if not client.get_ioption("quiet"):
        client.output(aliasname+" now aliases "+aliastext+"\n")
    #snl()

def unalias(client, text):
    """    removes an alias
    usage: #unalias <shortcut> 
    """

    if client.aliases.has_key(text):
        del client.aliases[text]
        if not client.get_ioption("quiet"):
            client.output(text+" unaliased.\n")
    else:
        client.output("No alias with the name %s defined.\n" % text)
    #snl()
    

    
def listaliases(client, text):
    for a in client.aliases.keys():
        client.output(a+" aliases to "+client.aliases.get(a))
    snl()

def rem_trigger(client, text):
    """    removes a trigger
    usage: #untrigger <trigger text>|<trigger nr>
    """
    if len(text)<=0:
        client.output("Remove which trigger?")
    else:
        if client.triggers.has_key(text):
            del client.triggers[text]
            if not client.get_ioptions("quiet"):
                client.output("Remove trigger on:"+text)
        else:
            nr = -1
            try:
                nr = int(text)
            except:
                client.output("No such trigger found.")
                pass
            if nr >= 0:
                keys = client.triggers.keys()
                keys.sort()
                n = 1
                for t in keys:
                    if n==nr:
                        del client.triggers[t]
                        if not client.get_ioptions("quiet"):
                            client.output("Removed trigger nr."+str(n)+")")
                        return
                    n = n + 1
            client.output("No such trigger found")                
                
def set_trigger(client, text):
    """    adds a trigger

    usage: #trigger <arbitrary text to trigger on>~<commands to be sent>
           As with aliases, commands can be separated by '\\;'.
           See help alias for an example. Instead of text, a regexp can
           be used, too. Matching is case sensitive.
    """
    
    if len(text)<1:
        if len(client.triggers.keys())>0:
            keys = client.triggers.keys()
            keys.sort()
            nr = 1
            for t in keys:
                client.output("%d) %s triggers %s" % (nr,t,client.triggers[t][1]))
                nr = nr + 1
        else:
            client.output("No triggers defined.")

        
    else:
        (what,action)=string.split(text,"~",1)
        client.triggers[what] = (re.compile(what), action)
        if not client.get_ioption("quiet"):
            client.output(what+" now triggers "+action+".")


def write_file(client, params):
    """    writes all (alias,triggers, options, etc.) definitions to a single file
    usage: #write <filename>
           The file can be re-read by the #read command.
    """
    client.write_file(params)

def set_font(client, params):
    """ sets the font for the terminal window of the gui (gui)
    usage: #font <fontname>~<size>~<bold|normal>
    """
    if params:
       param=params.split('~')
       client.gui.setfont((param[0], int(param[1]), param[2]))
           
def set_color(client, params):
    """ sets a color of the terminal window to a new value (gui)
    usage: #font <colorname>~<colorstring>
    The colorstring is a Tk colorstring, hexadecimal, with
    two bytes each for red, green, blue. Possible colors are:
    black, red, green, blue, cyan, magenta, yellow. The bright
    variants are written as bright black, bright red, etc.
    'foreground' and 'background' entries can be used for the
    primary colors.
    """
    if params:
        param=params.split('~')
        client.gui.set_color(param[0],param[1])
    
def save_trigger(client, file):
  for k in client.triggers.keys():
    file.write("#trigger %s~%s\n" % (k,re.sub(r";",r"\;",client.triggers[k][1])))

def save_affects(client, file):
    for a in client.get_affects().values():
        if not a.standard:
            if a.fading:
                #affect <name>~<symbol>~<affect>~<spellup>~<spelldown>[~<fade>]
                file.write("#affect %s~%s~%s~%s~%s~%s"%(a.name,ansitostring(a.symbol),a.aff_re_orig,a.spellup.pattern,a.spelldown.pattern,a.fading.pattern))
            else:
               file.write("#affect %s~%s~%s~%s~%s"%(a.name,ansitostring(a.symbol),a.aff_re_orig,a.spellup.pattern,a.spelldown.pattern)) 
                                                     
def save_aliases(client, file):
   for k in client.aliases.keys():
     file.write("#alias %s %s\n" % (k,re.sub(r";",r"\;",client.aliases[k])))

def save_highlites(client, file):
  for h in client.highlights:
    file.write("#highlight %s %s\n" %(h.colorspec, h.trigger_re))
    
def save_buttons(client, file):
    if client.gui:
        for b in client.gui.affect_buttons:
            file.write("#defbutton %s~%s\n"%(b.name,b.command))
            if not client.get_affects().has_key(b.name):
                file.write("#buttonfg %s~%s\n"%(b.name,b.cget("foreground")))
                file.write("#buttonbg %s~%s\n"%(b.name,b.cget("background")))
        
def save_options(client,file):
    if client.gui:
        client.gui.set_sizeoptions()
        client.gui.set_options()
        file.write("#font %s~%i~%s\n"%client.gui.fontname)
        client.gui.write_colors(file)
        
    for k in client.options.keys():
        file.write("#setoption %s~%s\n" %(k,client.options[k]))
    if client.echo_commands:
        file.write("#echoinput on\n")
    else:
        file.write("#echoinput off\n")
    if splitscreen:
        file.write("#split\n")
    else:
        file.write("#unsplit\n")
    if client.triggers_on:
        file.write("#ignore off\n")
    else:
        file.write("#ignore on\n")
    if client.log:
        file.write("#log "+client.log.name+"\n")
    else:
        file.write("#log\n")
    file.write("#prefix ");
    if client.prefix:
        file.write(client.prefix)
    file.write("\n")
    
    
save_functions=[save_aliases,save_trigger,save_highlites,save_buttons,save_options,save_affects]

def write_affects(client,params):
    client.write_affects(params)
    
def write_triggers(client, params):
    """    writes the trigger definitions to a file
    usage: #writetriggers <filename>
           The file can be re-read with #read
    """
    client.write_triggers(params)
  
def write_aliases(client, params):
    """    writes the alias definitions to a file
    usage: #writealiases <filename>
           The file can be re-read with #read 
    """
    client.write_aliases(params)

def write_options(client, params):
    
    """   writes the preferences and gui-options to a file
    usage: #writeoptions <filename>
    """
    client.write_options(params)
def write_buttons(client, params):
    """ writes the button definitions to a file
    usage: #writebuttons <filename>
    """
    client.write_buttons(params)
    
def write_highlites(client, params):
    """    writes the highlight definitions to a file
    usage: #writehighlites <filename>
           The file can be re-read with #read
    """
    p = string.strip(params)
    if len(p) > 0:
        file = open(params,"w")
        save_highlites(client,file)
        if not client.get_ioption("quiet"):
            client.output("Highlites saved to"+p)

def write_speedwalks(client, params):
    """    writes the speedwalk definitions to a file
    usage #writetriggers <filename>
    """
    client.write_speedwalks(params)
    



def toggle_trigger(client, params):
    """    sets/switches triggering
    usage #ignore
            toggles triggering
          #ignore [on|off]
            sets ignoring triggers to on/off
    """
    param=string.strip(params)
    if len(param) > 0:
        if param == "on":
            client.triggers_on=0
        else:
            client.triggers_on=1
    else:
        client.triggers_on = not clienttriggers_on
    if client.triggers_on:
        if not client.get_ioption("quiet"):
            client.output("triggers are now active")
    else:
        if not client.get_ioption("quiet"):
            client.output("triggers are ignored from now on")

def com_eval(client, params):
    """    evaluates an arbitrary python expression

    usage: #eval <expression>
    sends <expression> to python for evaluation.
    """
    #client.output("exec(\""+params+"\")")
    exec params in globals()
    ##if erg != None:
##        print erg


def do_nothing(client, params):
    pass

def allprompt_command(client, params):
    """    sets a new prompt and battleprompt, using the same parameters

    usage: #allprompt <promptstring>
    Only if this command is used to set both, the prompt and the battleprompt, the applets will function properly.
    """
    send(client, "prompt "+params+infograbber_prompt)
    send(client, "battleprompt "+params+infograbber_prompt)


def prompt_command(client, params):
    """    sets a new prompt

    usage: #prompt <promptstring>
    Only if this command is used to set the prompt, the applets will function properly.
    """
    send(client, "prompt "+params+infograbber_prompt)

def battleprompt_command(client, params):
    """    sets a new battleprompt

    usage: #battleprompt <promptstring>
    Only if this command is used to set the battleprompt, the applets will function properly.
    """
    send(client, "battleprompt "+params+infograbber_prompt)



def connect(client, params):
    """    connects to a foreign host

    usage: #connect [hostname] [port]
    where hostname is the name of the host you want to connect
    to and port is a positive integer"""
    
    global t

    if t:
        client.output("Already connected to %s:%s" % (t.host,t.port))
    else:
        host = "dawn.rplay.net"
        port = 8000
        args = string.split(params)
        if len(args)>0:
            host = args[0]
        if len(args)>1:
            port = int(args[1])
        t = telnetlib.Telnet(host, port)
        thread.start_new_thread(client.recfun,())
        #thread.start_new_thread(client.recfun_new,())

def disconnect(client, params):
    """    disconnects from a foreign host

    usage: #disconnect

    NOTE: your Char is not saved or quitted!"""
    
    global t
    if t:
        client.output("Closing connection to %s:%s " % (t.host,t.port))
        t = None   ## fun running in thread checks for t!=None to continue -ph.
    else:
        client.output("Not connected!")
        
def help(client, params):
    """    gives help about the built-in commands
    usage: #help [topic]
           if no topic is given, all commands are listed, otherwise
           the help to the given command is printed.
    """
    if len(params)>0:
        if params[0]!="#":
            params = "#"+params
        if (client.commands.has_key(params)):
            client.output(params+":")
            if client.commands[params].__doc__:
                client.output(client.commands[params].__doc__)
        else:
            client.output("Unknown command.\n")
    else:
        client.output("The following commands are implemented:\n")
        keys = client.commands.keys()
        keys.sort()
        for k in keys:
            if k[0] != "~":
                doc = ""
                if client.commands[k].__doc__:
                    lines = string.split(client.commands[k].__doc__,"\n")
                    doc = lines[0]
                client.output(k+" "+doc)
        client.output(" ")
    
def com_exit(client, params):
    """    exits the mud client"""
    global done
    done = 1
    print "Goodbye, hope to see you again"
    if tk:
        tk.quit()
    client.cancel_delay()
    
def setvar(client, params):
    """    sets the value of a variable
    usage:  #setvar variable value
    The value can be retrieved in #eval-statements, or with
    $[variable]
    """
    plist = string.split(params," ",1)
    quiet = client.get_ioption("quiet")
    while plist[0][0] == "-":
        if plist[0][1] == "quiet":
            quiet = 1
        if plist[0][1] == "echo":
            quiet = None
        plist = plist[1:]
    if len(plist)== 2 and len(plist[1])>0:
        #(var, value) = string.split(params," ",1)
        statement = "%s = %s" % (plist[0],plist[1])
    else:
        statement = r'%s = ""' % plist[0]
    #client.output("statement:%s" % statement)
    #print "statement:",statement
    try:
        exec statement in globals()
        #client.output("Variable %s set to %s\n"%(var,value))
        #client.output("Variable %s set to %s\n"%(plist[0],plist[1]))
        if not quiet:
            client.output("Setting variable %s" % params)
    except:
        #client.output("Error setting %s to %s\n"%(plist[,value))
        client.output("Error setting %s "% params ) #(plist[,value))
        pass


def load_file(client, name):
    """    loads the named file with python code
    usage: #load <filename>
           The code in the file is interpreted by python,
           as if it was code within the client-file.
    """
    execfile(name,globals())
    if not client.get_ioption("quiet"):
        if client:
            client.output("Loaded file: "+name)
        else:
            print "Loaded file: "+name

def repeat_last(client, params=None):
    """    repeats the last command entered
    usage:  !
    """
    client.process_command(last_line)

def sleep(client, param=1):
    """ Lets the client sleep a bit
    usage: #sleep <seconds>
           The number of seconds can also be given as a fraction.
    """
    if param:
        try:
            f=float(param)
            if not client.get_ioption("quiet"):
                client.output("Sleeping "+param+"seconds.\n")
            time.sleep(f)
        except:
            client.output("Invalid number of seconds: "+param+"\n")
    else:
        client.output("You must provide a number of seconds to sleep!\n")

def shell(client, params):
    """    executes a shell command
    usage: #shell <commands to execute>
    """
    f = os.popen(params,"r")
    erg = f.readlines()
    f.close()
    if not client.get_ioption("quiet"):
        client.output(string.join(erg,""))
    
def split(client, params=None):
    """      Splits the screen into input- and output window (terminal)
    usage: #split
    A split scren is necessary for the applets to show
    """
    global splitscreen
    #    sys.stdout.write(clearup)

    
    if client and client.gui:
        return
    
    sys.stdout.write("\33[r")  # reset scrolling region

    sys.stdout.write("\33[2J\33[1;%dr\33[%d;1f" % (split_line-1, split_line))
    if tf_look:
        sys.stdout.write("_________Dawn______________________________________________________________");
    else:
        line = "["+"-"*(width-2)+"]"
        line = " "*width
        sys.stdout.write(statuscolor)
        for i in range(status_size):
            sys.stdout.write(line)


    #sys.stdout.write("\33[%d;1f\337" % (split_line-1))
    sys.stdout.write("\33[%d;1f" % (split_line+status_size));
    splitscreen=1


def unsplit(client, params):
    """ Removes splitting/applets (terminal)
    usage: #unsplit
    """
    global splitscreen
    if client.gui:
        return

    #sys.stdout.write("\33c\33[2J")
    sys.stdout.write("\33[r")  # reset scrolling region
    splitscreen = 0


def show_clock(client, params):
    """ shows the clock window"""
    if client.gui:
        
        client.clock = Clock.Clock()
        t = client.get_var("game_time")
        if t:
            client.clock.adjust_after(int(t.split()[0]))
            
    else:
        client.output("Clock only in GUI mode.")
        
    

def create_win(client, params):
    """    creates a new text output window
    usage: #createwin windowname windowtitle
    windowname is the name to use with the wecho command
    windowtitle is an optional title to display into the window
    works only with gui
    """
    if client.gui:
        args = string.split(params," ",1)
        name = None
        title = None
        
        if len(args)< 1:
            return
        if len(args)>0:
            name = args[0]
        if len(args)>1:
            title = args[1]

        if title:
            win = Term.TermOutput(client, name, title,transient=1)
        else:
            win = Term.TermOutput(client, name, transient=1)


        client.add_win(win)

def remove_win(client, params):
    """    removes a text output window
    usage: #removewin windowname
    works only with gui
    """
    if client.gui:
        client.remove_win(params)

def clear_win(client, params):
    """    clears the text in the given window
    usage: #clearwin <name>
    works only with gui
    """
    if client.gui:
        client.clear_win(params)

def wecho(client, params):
    """    sends text to a given window
    usage: #wecho name text
    works only with gui
    """
    if client.gui:
        args = string.split(params, " ", 1)
        name = args[0]
        text = ""
        if len(args)>1:
            text = args[1]
        client.send_win(name,text+"\n")
    
def echo(client, params):
    """    outputs some text to the screen
    usage: #echo <text>
    just displays <text> without anything else happening
    """
    client.output(params)

def last_lines(client, param):
    """  output the last <param> lines from the recall buffer
    """

    try:
        n = int(param)
    except:
        client.output("Invalid number!\n",0)
        return
    l = len(recall_buffer)
    s = max(1,l-n)
    for i in range(s,l):
        client.output(recall_buffer[i],0)
		

def recall_from_buffer(client, params):
    """   outputs all lines in the recall buffer that match the regexp given.
    usage: #recall [-i] [-c <lines>] <regexp>
          Parameters:
          -i: Ignore case
          -c: Also output the <lines> lines of context around the match.
          
          The match doesn't have to occur at the beginning of a line.
    """
    if params:
        p=string.split(params," ",1)
        flags = 0
        context = 0
        while p[0][0] == "-":
            if p[0][1] == "i":
                flags |= re.I
                p = string.split(p[1]," ",1)
            else:
                if p[0][1] == "c":
                    p=string.split(p[1]," ",1)
                    context=int(p[0])
                    p=string.split(p[1]," ",1)
            params = p[0]
        if (flags != 0):
            reg = re.compile(params,flags)
        else:
            reg = re.compile(params,flags)
        client.output("begin recall:\n")
        for i in range(0,len(recall_buffer)-1):
            if reg.search(recall_buffer[i]):
                start = max(0,i-context)
                end = min(len(recall_buffer),i+context)
                for line in recall_buffer[start:end+1]:
                    client.output(line,0)
        client.output("end recall\n")
    
def set_recall_size(client, param):
    """     Sets the size of the recall buffer.
    usage: #recallsize <lines>
           The standard number of lines for the recall buffer is 10000
    """
    global recall_buffer
    client.recall_size = int(param)
    if client.recall_size <0 :
        client.recall_size = 0
    recall_buffer=recall_buffer[-client.recall_size:]
    if not client.get_ioption("quiet"):
        client.output("New size of recall buffer: %i\n"%client.recall_size)

def write_recall(client, params):
    """     Writes the lines in the input buffer to a file (to keep as a log, eg.)
    usage: #writerecall <filename>
           This can be useful if something unexpected happened and you had
           logging turned off. If it's still in the recall-buffer, you can
           simply do a #writerecall to archive it.
    """
    global recall_buffer
    p = string.strip(params)
    if len(p)>0:
        file = open(p,"w")
        for line in recall_buffer:
            file.write(line+"\n")
        file.close()
        if not client.get_ioption("quiet"):
            client.output("Wrote recall-buffer to "+p+"\n")
    else:
        client.output("No filename given.\n")

def highl(client, text):
    """    defines an highlight or lists the defined ones
    usage #highlight <color spec> <regexp for the highlight expression>
          Just use the color's name as the color spec.
    """
    
    if len(text)<1:
        client.output("The following highlights are defined:")
        for h in highlights:
            client.output(h.trigger_re+" is highlighted in "+h.colorspec)
        if len(highlights)<1:
            client.output("None.")
        #snl()
        return
    (color,regexp)= string.split(text," ",1)
    highlights.append(HighLight(regexp,color))
    if not client.get_ioption("quiet"):
        client.output(regexp+" is highlighted in "+color+"\n")

def unhighl(client, text):
    """    removes a defined highlight
    """

    for i in range(len(highlights)):
        h = highlights[i]
        if h.trigger_re == text:
            del highlights[i]
            if not client.get_ioption("quiet"):
                client.output(text+" is no longer highlighted.")
            return
    client.output("No highlight found on "+text)

def do_log(client, param):
    """    starts/stops logging
    usage: #log [filename]
    if filename is given, starts logging into the named file (append),
    else logging is stopped.
    """
    if (not client.log) and (not param):
        client.output("Logging is stopped already - please provide a filename")
    if client.log:
        client.log.close()
        client.log = None
        if not client.get_ioption("quiet"):
            client.output("logging stopped\n")
    if param:
        param = string.strip(param)
        if len(param) > 0:
            client.log = open(param,"a")
            if not client.get_ioption("quiet"):
                client.output("logging to file: "+param+"\n")
    
        

def do_echo_input(client, param):
    """   toggle/set input echo
    usage: #echoinput
              toggles input echo
           #echoinput [on|off]
              sets input echo """
    #global echo_commands
    param=string.strip(param)
    if len(param) > 0:
        if param == "on":
            client.echo_commands=1
        else:
            client.echo_commands=0
    else:
        client.echo_commands = not client.echo_commands
    if client.echo_commands:
        if not client.get_ioption("quiet"):
            client.output("input echo on\n")
    else:
        if not client.get_ioption("quiet"):
            client.output("input echo off\n")
    
def do_prefix(client, param):
    """   sets the prefix command"""
    client.prefix = param
    if client.prefix:
        if not client.get_ioption("quiet"):
            client.output("Prefix for all commands is now %s\n" % client.prefix)
    else:
        if not client.get_ioption("quiet"):
            client.output("Prefix removed.\n")

def show_gui(client, params):
    """ Displays the gui and button-bar (gui)
    usage: #showgui
    Standard options for the gui are read from std.opt.
    """
    if not have_tkinter:
        client.output("Tkinter module needed for gui.")
        return
    
    global gui
    global tk

    tk = Tk()
    gui=Gui.GUI(client, tk)
    client.gui = gui

    if not client.get_ioption("quiet"):
        client.output("Loading gui-settings from std.opt")
    read_file(client, "std.opt")

def show_affects(client, params):
    """Opens a new terminal showing affects
    usage: #showaffects
    """
    open_fifo(1,"Affects")
    if not client.get_ioption("quiet"):
        client.output("Showing affects now\n")

def close_affects(client, params):
    """Closes the affects window
    usage: #closeaffects
    """
    if fifo[1]:
        close_fifo(1)
        if not client.get_ioption("quiet"):
            client.output("No longer showing affects\n")
    else:
        client.output("Affect window already closed\n")

def title(client, text):
    """Displays text as title of your terminal window, if its xterm or compatible."""
    client.title = text
    if client.gui:
        client.gui.master.title(text)
    else:
        print "\33]0;%s\7" % text,
 

def defkey(client, text):
    """Defines a global key for the gui. (gui)
       usage: defkey keycode text-to-send
    """
    (code, param) = string.split(text," ",1)
    client.defkey(code, param)
    
        
def set_option(client,param):
    """   Sets an option for the gui
    usage: #setoption option~value
    """
    (name,value)=string.split(param,"~",1)
    client.set_option(name,value)

def read_file(client, text):
    client.read_file(text)
    
def send(client, text):
    client.send(text)

def addcommand(client, commandstring, command):
    client.commands[commandstring] = command

def is_separated(text):
    """return if text contains unescaped ';' to separate commands"""
    pos = string.find(text,";")
    while pos>=0:
        if pos == 0:
            return  1
        if text[pos-1] != "\\":
            return  1
        pos = string.find(text,";",pos+1)
    return 0

def separated_parts(text):
    """partition text by ;, but not by \;"""

    erg = []                            # list to take the parts
    pos = string.find(text,";")         # find first ;

    while pos>=0:                       # while found
        if pos == 0:                    # string starts with ; so just take the rest
            text = text[pos+1:]
            pos = 0
        else:                           # ; somewhere in the string
            if text[pos-1] != "\\":     # if not escaped by backslash
                erg.append(text[:pos])  # partition here
                text = text[pos+1:]
                pos = 0
            else:                       # set search position behind the ;
                pos = pos + 1 
        pos = string.find(text,";",pos) # and search the next occurrence of ;
        
    if len(text)>0:                     # if there is something left
        erg.append(text)                # append it
    return erg
    
def get_input():
    acc = ""
    line = raw_input()
    ##print "line:"+line+"<="
    acc = line
    if len(acc)> 0 and acc[-1]=="\\":
        acc = acc[:-1]
    while len(line)>0 and line[-1]=="\\":
        line = raw_input()
        ##print "line:"+line+"<="
        acc = acc + "\n"+ line
        if acc[-1]=="\\":
            acc = acc[:-1]
    ##    print "get_input:",acc,"<="
    return acc

def smash_escapes(text):
    """ remove all ansi escape sequences from text """
    text= string.replace(text,"\r", "") ## change \r\n to \n
    pos = string.find(text,"\33")
    while pos >= 0:
        pos2 = string.find(text,"m",pos)
        if pos2 > 0:
            text = text[:pos]+text[pos2+1:]
        else:
            text = text[pos+1:]
        pos = string.find(text,"\33")                       
    return text



def ter_resize(*args):
    global width, height,split_line
#    try:
    filenr = sys.stdin.fileno()

    if have_fcntl:
        arg = struct.pack("HHHH",0,0,0,0)
        arg = fcntl.ioctl(filenr, 0x5413, arg)
        #arg = fcntl.ioctl(filenr, termios.TIOCGWINSZ, arg)
        dims = struct.unpack("HHHH", arg)
        width = dims[1]
        height = dims[0]
        #print "tersize:%sx%s" % (width,height)
        split_line = height - input_size - status_size + 1
        if splitscreen:
            split(None)
#    except:
#        pass
    
if have_signal and have_fcntl:
    try:
        signal.signal(signal.SIGWINCH, ter_resize)
    except:
        pass

tk= None

class Mycompleter:
    def __init__(self):
        self.il={}
        
    def complete(self, text, state):
        self.matches=self.get_matches(text)
        try:
            return self.matches[state]
        except:
            return None
       
    def get_matches(self, text):
        mat=[]
        n=len(text)
        if (n==0):
            return mat
        
        for line in self.il.keys():
            try:
                if line[:n]==text:
                    mat.append(line)
            except:
                pass
        return mat
    
class DelayThread(threading.Thread):
    def __init__(self,time,commands,client):
        threading.Thread.__init__(self)
        self.time=time
        self.commands=commands
        self.client=client
        
    def run(self):
        s = sched.scheduler(time.time,time.sleep)
        self.scheduler = s
        self.events = []
        i=1
        for c in self.commands:
            self.events.append(s.enter(self.time*i/1000,1,self.client.process_command,[c]))
            i = i+1
        s.run()

    def stop(self):
        for e in self.events:
            try:
                self.scheduler.cancel(e)
            except:
                pass

class CDBEntry:
    def __init__(self, dname, pname, pw):
        self.dname = dname
        self.pname = pname
        self.pw = pw
        
class Client:
    def __init__(self):
        global client
        client = self     # Uhhggg....
	self.gui = None        
        self.root = None
        self.completer = Mycompleter()
        self.gui = None
        self.aliases = {}
        self.triggers = {}
        self.highlights = highlights
        self.speedwalks = {"tour" : "nee:buy beer:wwseeessssen:buy tea:drink tea:sessss"}
        self.triggers_on = 1
        self.echo_commands = 0
        self.log = None
        self.prefix = None
        self.recall_size=10000
        self.options = {}
        self.buttons = {}
        self.keybindings = []   # key definitions for gui
        self.queue = []
        self.delays = []
        self.variables = {}
        self.chardb = {}
        self.windows = {}
        self.time_benchmark = None #if true, timing information is sent to stdout
        self.login_name = None
        self.login_pw = None
        self.title="DMC"
        self.clock=None
        self.sre=re.compile(r'\[(.*)->(.*)', re.DOTALL)
        
        self.commands = {"#" : do_nothing,
                         "#alias"  : defalias,
                         "#trigger": set_trigger,
                         "#ignore" : toggle_trigger,
                         "#write": write_file,
                         "#writealiases": write_aliases,
                         "#writetriggers":write_triggers,
                         "#writehighlites":write_highlites,
                         "#writebuttons":write_buttons,
                         "#writeaffects":write_affects,
                         "#writeoptions":write_options,
                         "#title" : title,
                         "#connect": connect,
                         "#disconnect": disconnect,
                         "#help"   : help,
                         "#echoinput": do_echo_input,
                         "#createwin": create_win,
                         "#clock": show_clock,
                         "#removewin": remove_win,
                         "#clearwin": clear_win,
                         "#wecho": wecho,
                         "#echo"   : echo,
                         "#eval"   : com_eval,
                         "#setvar" : setvar,
                         "#defkey" : defkey,
                         "#split"  : split,
                         "#prefix" : do_prefix,
                         "#showaffects" : show_affects,
                         "#showlog" : lambda x: open_fifo(2,"Log"),
                         "#gui":show_gui,
                         "#closelog": lambda x: close_fifo(2),
                         "#closeaffects" : close_affects,
                         "#unsplit": unsplit,
                         "#shell": shell,
                         "#sleep": sleep,
                         "#send": send,
                         "#unalias": unalias,
                         "#prompt": prompt_command,
                         "#battleprompt": battleprompt_command,
                         "#allprompt": allprompt_command,
                         "#load": load_file,
                         "#read": read_file,
                         "#recall": recall_from_buffer,
                         "#recallsize": set_recall_size,
                         "#writerecall": write_recall,
                         "#log": do_log,
                         "#more": do_more,
			 "#last": last_lines,
                         "#untrigger": rem_trigger,
                         "#highlight": highl,
                         "#unhighlight": unhighl,
                         "#setoption":set_option,
                         "#font":set_font,
                         "#color":set_color,
                         "#defbutton":defbutton,
                         "#rembutton":rembutton,
                         "#repeat":dorepeat,
                         "#affect":defaffect,
                         "#delaffect":delaffect,
                         "#more_next": tab_pressed,
                         "~tab_pressed": tab_pressed,
                         "#delay":do_delay,
                         "#break":do_break,
                         "#jobs":list_jobs,
                         "#buttonbg":buttonbg,
                         "#buttonfg":buttonfg,
                         "#buttoncmd":buttoncmd,
                         "#buttoncolor":buttoncolor,
                         "#if":doif,
##            "!"       : repeat_last,
                         "#exit"   : com_exit}
        
    def get_var(self, name):
        if self.variables.has_key(name):
            return self.variables[name]
        return None

    def set_var(self, name, value):
        self.variables[name] = value
        
    def append_queue(self,new_commands):
        s = len(self.queue)
        if type(new_commands) is types.ListType:
            for c in new_commands:
                self.queue.append(c)
        elif is_separated(new_commands):
            for c in separated_parts(new_commands):
                self.queue.append(c)
        else:
            self.queue.append(new_commands)
        if s == 0:
            self.process_command(self.queue.pop(0))
    
    def clear_queue(self):
        self.queue=[]

    def cancel_delay(self,number=None):
        if number:
            self.delays[number].stop()
        else:
            for t in self.delays:
                t.stop()
                
    def delay(self,time,commands):
        if type(commands) is types.ListType:
            th=DelayThread(time,commands,self)
        elif is_separated(commands):
            th=DelayThread(time,separated_parts(commands),self)
        else:
            th=DelayThread(time,[commands],self)
        self.delays.append(th)
        th.start()
        return len(self.delays)-1

    def add_win(self, win):
        self.remove_win(win.name)      # remove any existing window with that name
        self.windows[win.name]=win

    def remove_win(self, name):
        if self.windows.has_key(name):
            win = self.windows[name]
            win.withdraw()
            del self.windows[name]

    def send_win(self, name, text):
        if self.windows.has_key(name):
            self.windows[name].output(text)

    def clear_win(self, name):
        if self.windows.has_key(name):
            self.windows[name].clear()
            
            
    def resetaffects(self):
        global affects
        affects.clear()
        for a in std_affects:

            if a.fading:
                affects[a.name]=(Client.Affect(a.name,a.symbol,a.aff_re_orig,a.spellup.pattern,a.spelldown.pattern,a.fading)) 
            else:
                affects[a.name]=(Client.Affect(a.name,a.symbol,a.aff_re_orig,a.spellup.pattern,a.spelldown.pattern))

    def set_option(self,name,value):
        self.options[name]=value
#        if self.gui:
#            self.gui.default_windowpos()
    
    def get_option(self, name, default = None):
        """ get option stored to the key
        returns default value, which defaults to None"""
        
        if self.options.has_key(name):
            return self.options[name]
        return default

    def get_ioption(self, name, default = 0):
        """get an option as an integer value or 0"""
        try: 
            return int(self.options[name])
        except:
            pass
        return default

    

    def load_files(self):

        command_path = args[0] # use argv[0]
        applet_dir = os.path.dirname(command_path) # cut off the client.py part

        command_dir = applet_dir
        self.output("command_dir: "+command_dir)
        self.output("applet_dir : "+applet_dir)

        if os.access(applet_dir+"/Applets.py",os.F_OK):
            load_file(self, applet_dir+"/Applets.py")
            self.output("\n")
        elif os.access("Applets.py",os.F_OK):
            load_file(self, "Applets.py")
        else:
            self.output("Applets.py not found\n")

        homedir = ""

        # set command dir here, so it can be overridden by ".clientrc"
        # default value is the directory where the client.py is located   

        if os.environ.has_key("HOME"):
            homedir = os.environ["HOME"]
        if os.access(homedir+"/.clientrc",os.F_OK):
            load_file(self, homedir+"/.clientrc")
            #self.output("read config file:"+homedir+"/.clientrc\n")
        elif os.access(".clientrc",os.F_OK):
            load_file(self, ".clientrc")
            #self.output("reading default config file:"+"/.clientrc\n")
            #print "reading default config file:"+"/.clientrc\n"
        else:
            self.output("no configfile found\n")
            #print "no configfile found"

        # if available, check for personal preferenced files,
        # else load default ones.

        if os.access(command_dir+"/speedwalk.py",os.F_OK):
            load_file(self, command_dir+"/speedwalk.py")
        elif os.access("speedwalk.py",os.F_OK):
            load_file(self, "speedwalk.py")
        else:
            self.output("speedwalk.py not found\n")

        if os.access(command_dir+"/spells.py",os.F_OK):
            load_file(self, command_dir+"/spells.py")
        elif os.access("spells.py",os.F_OK):
            load_file(self, "spells.py")
        else:
            self.output("spells.py not found\n")

        if tf_look:
            load_file(self, "TF.py")

        if len(args)>1:
            read_file(self,args[1])

        self.read_chardb()

    def init_readline(self):
        if have_readline:
            try:
                readline.set_completer_delims("")
                readline.set_completer() 
                readline.set_completer(self.completer.complete)
                readline.parse_and_bind("tab: complete")

                readline.parse_and_bind('"\e[11~": "\C-a\C-kF1\n\C-y"')
                readline.parse_and_bind('"\e[12~": "\C-a\C-kF2\n\C-y"')
                readline.parse_and_bind('"\e[13~": "\C-a\C-kF3\n\C-y"')
                readline.parse_and_bind('"\e[14~": "\C-a\C-kF4\n\C-y"')
                readline.parse_and_bind('"\e[15~": "\C-a\C-kF5\n\C-y"')
                readline.parse_and_bind('"\e[17~": "\C-a\C-kF6\n\C-y"')
                readline.parse_and_bind('"\e[18~": "\C-a\C-kF7\n\C-y"')
                readline.parse_and_bind('"\e[19~": "\C-a\C-kF8\n\C-y"')
                readline.parse_and_bind('"\e[20~": "\C-a\C-kF9\n\C-y"')
                readline.parse_and_bind('"\e[21~": "\C-a\C-kF10\n\C-y"')
                readline.parse_and_bind('"\e[23~": "\C-a\C-kF11\n\C-y"')
                readline.parse_and_bind('"\e[24~": "\C-a\C-kF12\n\C-y"')
            except:
                self.output("using old version of readline")

    def run_gui(self):
        show_gui(self, "")
        self.load_files()
        self.do_start()
        for b in self.keybindings:
            b.bind()
            
        tk.mainloop()
        
    def run_terminal(self):
        ter_resize()
        self.load_files()
        self.init_readline()
        self.do_start()
        while 1:
            if done:
                break
            command = get_input()
            if splitscreen:
                sys.stdout.write("\33[%d;1H\33[0J" % (split_line+status_size))

            self.process_command(command)
            last_line = command
            self.completer.il[command]=command


        self.do_exit()

    def do_start(self):
        self.output("""

Mud Client  Copyright (C) 2002 Peter Herth <p.herth@uni-koeln.de>

enter #help for a list of commands available.

"""+fcolor(RED,bold=1)+"""*** IMPORTANT ***"""+clr()+""": for a faster update and a proper display of the status bar,
                   redefine your mud prompt via:
  
                   #prompt yourpreferredpromptoptions
                   #battleprompt yourpreferredpromptoptions

		   OR (if prompt and battleprompt should look alike: )

                   #allprompt yourpreferredpromptoptions

""")
        #print "charlist:",self.charlist()

    def do_exit(self):
        for i in range(len(fifo)):
            close_fifo(i)
        for t in self.delays:
            t.stop()
        sys.stdout.write(clearup)
        print "Goodbye, hope to see you again"

    def process_command(self, text):
        #print 'process_command("%s")' % text
        if is_separated(text):
            erg = separated_parts(text)
            for c in erg:
                self.process_command(c)
            return
    
        #print "text:",text,"r:",re.sub(r"\\;", ";", text)
        text = re.sub(r"\\;", ";", text)

        s = string.split(text," ",1)
        params =""
        if len(s)>1:
            params = s[1]
        command = s[0]
    
        if self.aliases.has_key(command):                 # alias expansion
            ##send_fifo(1, "expanding:"+command)
            repl = self.aliases.get(command)
            pos = string.find(repl,"$")              
        
            if pos >= 0:                             # if expansion string contains $ for parameters
                paramlist = string.split(params," ") # expand them 
                text = expand(repl, paramlist)
                self.process_command(text)
                return
            else:                                    # append whole parameter string to the expansion
                text = self.aliases.get(command)+" "+params
                ##send_fifo(1," => "+text+"\n")
                self.process_command(text)
                return

    
        if self.commands.has_key(command):
            fun = self.commands.get(command)
            if debug:
                fun(self, params)
            else:
                try:
                    if eval_brackets:
                        params = evaluate_brackets(params)
                    fun(self, params)
                except:
                    #traceback.print_exc(file=sys.stdout)
                    for s in traceback.format_exception(sys.exc_type, sys.exc_value, sys.exc_traceback):
                        self.output(string.rstrip(s))
                    self.output("")
        else:
            if eval_brackets:
                text = evaluate_brackets(text)

            if self.prefix:
                if not(text[0:6]=="#prefix"):
                    text = text.split("\n")
                    for c in text:
                        self.send(self.prefix+c)
            else:
                self.send(text)


    # write buffer to screen for output 
    def output(self, buffer,recall=1):
        global recall_buffer
        if self.log:
            self.log.write(buffer)
            self.log.write("\n")
            self.log.flush()
        if recall:
            recall_buffer.extend(string.split(buffer,"\n"))
        if len(recall_buffer)>self.recall_size:
            recall_buffer=recall_buffer[-self.recall_size:]
        if more:
            more_buffer.extend(string.split(buffer,"\n"))
            if self.gui:
                self.gui.master.title("%s [%d]"% (self.title, len(more_buffer)))
	    elif not splitscreen:
	        print CLR_TO_CURSOR,"\r(more %d)" % len(more_buffer),
	    return

        self.raw_output(buffer)

    def raw_output(self, buffer):
        if self.gui:
            gui.output("\n"+buffer)
        else:
            write_to_screen(buffer)


    def input_read(self, txt):
        global buffer
        global lasttime
        if (len(txt)>0):
            buffer = buffer + txt
        t = time.time()
        sys.stdout.flush()

        if (t-lasttime)>0.06 and len(buffer)>0:

            for a in applets:
                buffer = a.do_match(buffer)

            for a in affects.values():
                a.do_match(buffer)

            orig_buffer = buffer
            for g in gags:
                buffer = g.do_match(buffer)
            for g in highlights:
                buffer = g.do_match(buffer)
            update_statusbar()
            
            self.output(buffer)
            if self.gui:
                self.gui.notify_input()
            if len(self.queue) > 0:
                self.process_command(self.queue.pop(0))
                
            lasttime = t
            pos = 0

            if self.login_name:
                if name_prompt_re.search(orig_buffer):
                    #print "sending name:%s<=" % self.login_name
                    self.send(self.login_name)
                    
                if enter_re.search(orig_buffer):   # we are entering the mud -> forget name/password
                    #print "clearing pw"
                    self.login_name = None
                    self.login_pw = None
                    
            if self.login_pw:
                
                if password_prompt_re.search(orig_buffer):
                    #print "sending pw:%s<=" % self.login_pw
                    self.send(self.login_pw)
            
            orig_buffer = smash_escapes(orig_buffer)
            if self.triggers_on:
                for t in self.triggers.keys():
                    erg = self.triggers[t][0].findall(orig_buffer)
                    for e in erg:
                        self.process_command(expand(self.triggers[t][1],e))

            mg = self.sre.search(orig_buffer)
            
            if mg:
                try:
                    self.send_win(mg.group(1),mg.group(2))
                except:
                    pass
            buffer = ""
    
            update_statusbar()

    def recfun(self):
        global t
        erg = "asdf"
        while 1:
            erg = ""
            if not t:
                return
            try:
                if self.time_benchmark:
                    time.sleep(1)
                t1= time.time()
                erg = t.read_until("prompt>",0.3)
                if len(erg)>0:
                    t2= time.time()-t1
            except EOFError:
                break
            except:
                pass
                #traceback.print_exc(file=sys.stdout)
    
                
                
            l = len(erg)
            if l>7 and erg[l-7:]=="prompt>":
                erg = erg[:l-7]
            t3=time.time()
            self.input_read(erg)
            t4=time.time()
            if self.time_benchmark and len(erg)>0:
                print "%.10f    %.10f    %.10f" % (t4-t1,t4-t3,t2)
            
        self.output("\nConnection closed by foreign host.")
        t = None

    def recfun_new(self):
        global t
        erg = "asdf"
        while 1:
            erg = ""
            if not t:
                return
            try:
                #erg = t.read_some()
                erg = t.read_until("prompt>",0.3)
            except EOFError:
                break
            except:
                pass
                #traceback.print_exc(file=sys.stdout)
    
                
                
            l = len(erg)
            if l>7 and erg[l-7:]=="prompt>":
                erg = erg[:l-7]
            for line in string.split(erg,"\n"):
                if len(line)>0:
                    self.input_read(line)          
    
        self.output("\nConnection closed by foreign host.")
        t = None

    
    def send(self, text):
        if t:
            t.write(text+"\n")
            if self.echo_commands:
                self.output(echocolor+text+"\n")
            #print "sending ==>",text,"<=="
            #send_fifo(1, "send:"+text)

    def snl(self):
        self.send("")

    def render_affects(self):
        vals = actives.values()
        vals.sort()
        erg = ""
    
        i=0
        for v in vals:
            erg = erg + v
##            i = i+1
##            if i%10 == 0: 	     # 10 affects in any successive line
##                erg = erg + "\33[1B\33[50D"  # 1 line down and 50 (10*5) characters left
        return erg

    
    def read_file(self, filename):
        """     Reads the file as if typed into the client.
        usage: #read <filename>
               The files written by the #writexxx commands can be re-read
               by #read.
        """
        global eval_brackets
        if filename:
            eval_brackets = None
            try:
                file = open(filename)
                for line in file.readlines():
                    l = string.strip(line)
                    if len(l)>0:
                        self.process_command(l)
                file.close()
                if self.gui:
                    self.gui.use_options()
            except:
                self.output("Error reading file: "+filename)
            eval_brackets = 1
        else:
            self.output("You must provide a filename!\n")


    
    def write_file(self, params):
        """    writes all (alias,triggers, options, etc.) definitions to a single file
        usage: #write <filename>
               The file can be re-read by the #read command.
        """
        p = string.strip(params)
        if len(p)>0:
          file = open(p,"w")
          for f in save_functions:
            f(self, file)
          self.output("definitions saved to "+p)

    def write_triggers(self, params):
        p = string.strip(params)
        if len(p)>0:
            file = open(p,"w")
            save_trigger(self, file)
            self.output("\nTriggers saved to "+p)
    def write_affects(self, params):
        p = string.strip(params)
        if len(p) > 0:
            file = open(p,"w")
            save_affects(self,file)
            self.output("\nAffects saved to "+p)
            
    def write_aliases(self, params):
        p = string.strip(params)
        if len(p)>0:
            file = open(p,"w")
            save_aliases(self, file)
            self.output("\nAliases saved to "+p)

    def write_speedwalks(self, params):
        p = string.strip(params)
        if len(p)>0:
            file = open(p,"w")
            save_speedwalks(self, file)
            self.output("Speedwalks saved to "+p)
        
    def write_options(self,params):       
        p = string.strip(params)
        if len(p)>0:
            file = open(p,"w")
            save_options(self, file)
            self.output("Options saved to: "+p)
            
    def write_buttons(self, params):
        p = string.strip(params)
        if len(p)>0:
            file = open(p,"w")
            save_buttons(self, file)
            self.output("Buttons saved to: "+p)
            
    def add_gag(self, gag):
        gags.append(gag)

    def remove_gag(self, gag):
        try:
            gags.remove(gag)
        except:
            pass    

    def make_gag(self, trigger_re, alternate, a_exp,opts):
        return Gag(trigger_re, "", alternate, a_exp,opts)
    
    def get_affects(self):
        return affects
    
    def get_actives(self):
        return actives
    
    def set_affects(self,new_affects):
        affects.clear()
        for a in new_affects.values():
            affects[a.name]=a
            
    def get_fadings(self):
        return fadings

    def defkey(self, key, command,keycode=None):
        b = keybinding(self, key, command,keycode)
        self.keybindings.append(b)
        if self.gui:
            b.bind()


    def crypt(self, txt):
        erg = ""
        for c in txt:
            if c in string.lowercase:
                erg=erg+(chr(97+((ord(c)-97+13) % 26)))
            elif c in string.uppercase:
                erg=erg+(chr(65+((ord(c)-65+13) % 26)))
            else:
                erg=erg+c
        return erg
        

    def read_chardb(self):
        try:
            f = open(".dmc_chardb","r")
            for l in f.readlines():
                (dname, pname, pw)= l.strip().split("~")
                self.chardb[dname]=CDBEntry(dname, pname, self.crypt(pw))
            f.close()
        except:
            pass

    def add_to_chardb(self, dname, pname, pw, save=1):
        self.chardb[dname]=CDBEntry(dname, pname, pw)
        if save:
            try:
                f = open(".dmc_chardb","w")
                for k in self.chardb.keys():
                    e = self.chardb[k]
                    f.write("%s~%s~%s\n" % (e.dname, e.pname, self.crypt(e.pw)))
                f.close()
            except:
                pass
                
                           
    def do_connect(self, name, pw):
        self.login_name = name
        self.login_pw = pw
        #print "password=>%s<" % pw
        self.process_command("#connect")


    def autoconnect(self):
        self.login_name = self.get_option("login_name")
        self.login_pw = self.crypt(self.get_option("login_pw",""))
        if not self.login_name:
            self.output("No login name set!")
        elif not self.login_pw:
            self.output("No login password set!")
        else:
            self.process_command("#connect")

    def charlist(self):
        """ return a list of all .dmc files in the subdirectory characters,
        it is created on the fly, when it does not exist"""
        erg = []
        if not os.access("characters", os.F_OK):        
            self.output("Creating character directory.")            
            try:
                os.mkdir("characters")
            except:
                self.output("cannot create character directoy!")
                return erg

        files = os.listdir("characters")
        for f in files:
            if f[-4:]==".dmc":
                erg.append(f)
                
        return erg
